;;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;;
;;;; CMUCL 20a Patch 000: Fix issue with printing '|\|| and '|`|.
;;;; These symbols were printed without escaping.  This only applies
;;;; to the Unicode builds.
;;;;

(pushnew :cmucl-20a-patch-000 *features*)

(in-package "LISP")

;; Get rid of old reinit function
#+(and unicode cmu20a (not cmucl-20a-patch-000))
(setf ext::*after-save-initializations*
      (remove #'reinit-char-attributes ext:*after-save-initializations*))

#+(and unicode cmu20a (not cmucl-20a-patch-000))
(ext:without-package-locks
(defun reinit-char-attributes ()
  (unless (probe-file +unidata-path+)
    (cerror "Continue anyway" "Cannot find ~S, so unicode support is not available"
	    +unidata-path+)
    (return-from reinit-char-attributes nil))
  (flet ((set-bit (char bit)
	   (let ((code (char-code char)))
	     (setf (aref character-attributes code)
		   (logior bit (aref character-attributes code))))))

    (fill character-attributes 0)
    (dolist (char '(#\! #\@ #\$ #\% #\& #\* #\= #\~ #\[ #\] #\{ #\}
		    #\? #\< #\>))
      (set-bit char other-attribute))

    (dotimes (i 10)
      (set-bit (digit-char i) number-attribute))

    (do ((code (char-code #\A) (1+ code))
	 (end (char-code #\Z)))
	((> code end))
      (declare (fixnum code end))
      (set-bit (code-char code) uppercase-attribute)
      (set-bit (char-downcase (code-char code)) lowercase-attribute))

    (set-bit #\- sign-attribute)
    (set-bit #\+ sign-attribute)
    (set-bit #\^ extension-attribute)
    (set-bit #\_ extension-attribute)
    (set-bit #\. dot-attribute)
    (set-bit #\/ slash-attribute)
    (set-bit #\` funny-attribute)
    (set-bit #\| funny-attribute)

    ;; For everything else, derive the attribute from the Unicode
    ;; properties of the character.
    (dotimes (i char-code-limit)
      (when (zerop (aref character-attributes i))
	(let* ((char (code-char i)))
	  ;; What about surrogate pairs?  What should we do about them?
	  (cond ((upper-case-p char)
		 (set-bit char uppercase-attribute))
		((lower-case-p char)
		 (set-bit char lowercase-attribute))
		((or (= (unicode-category i) +unicode-category-other+)
		     (= (ash (unicode-category i) -4) 6))
		 ;; Includes characters without case and all Unicode symbols
		 (set-bit char othercase-attribute))
		(t
		 (setf (aref character-attributes i) funny-attribute))))))))
)

;; Put new version on the list
#+(and unicode cmu20a (not cmucl-20a-patch-000))
(pushnew #'reinit-char-attributes ext:*after-save-initializations*)

;; Update the table correctly
#+(and unicode cmu20a (not cmucl-20a-patch-000))
(reinit-char-attributes)
